#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Rewrite and/or check time axis of MIP NetCDF files.

"""

import logging
import os
import re
from netcdftime import datetime

import numpy as np
from constants import *
import db
from context import ProcessingContext
from handler import File
from nctime.utils.misc import trunc
from nctime.utils.time import time_inc


def process(collector_input):
    """
    time_axis_processing(inputs)

    Time axis process that\:
     * Deduces start and end dates from filename,
     * Rebuilds the theoretical time axis (using frequency, calendar, etc.),
     * Compares the theoretical time axis with the time axis from the file,
     * Compares the last theoretical date with the end date from the filename,
     * Checks if the expected time units keep unchanged,
     * Checks the squareness and the consistency of time boundaries,
     * Rewrites (with ``--write`` mode) the new time axis,
     * Computes the new checksum if modified,
     * Traceback the status.

    :param tuple collector_input: A tuple with the file path and the processing context
    :returns: The updated file handler instance
    :rtype: *nctime.axis.handler.File*

    """
    # Deserialize inputs from collector
    ffp, ctx = collector_input
    # Block to avoid program stop if a thread fails
    try:
        # Instantiate file handler
        fh = File(ffp=ffp, has_bounds=ctx.tinit.has_bounds)

        # Extract start and end dates from filename
        start, _, _ = fh.get_start_end_dates(pattern=ctx.pattern,
                                             frequency=ctx.tinit.frequency,
                                             units=ctx.tinit.funits,
                                             calendar=ctx.tinit.calendar)
        # Rebuild a theoretical time axis with high precision
        fh.time_axis_rebuilt = fh.build_time_axis(start=start,
                                                  inc=time_inc(ctx.tinit.frequency)[0],
                                                  input_units=ctx.tinit.funits,
                                                  output_units=ctx.tinit.tunits,
                                                  calendar=ctx.tinit.calendar,
                                                  is_instant=ctx.tinit.is_instant)
        # Check consistency between last time date and end date from filename
        if fh.last_timestamp != fh.end_timestamp:
            # Rebuild a theoretical time axis with low precision
            fh.time_axis_rebuilt = fh.build_time_axis(start=trunc(start, 5),
                                                      inc=time_inc(ctx.tinit.frequency)[0],
                                                      input_units=ctx.tinit.funits,
                                                      output_units=ctx.tinit.tunits,
                                                      calendar=ctx.tinit.calendar,
                                                      is_instant=ctx.tinit.is_instant)
            if fh.last_timestamp != fh.end_timestamp:
                fh.status.append('003')
            elif fh.last_date != fh.end_date:
                fh.status.append('008')
        # Check consistency between instant time and time boundaries
        if ctx.tinit.is_instant and ctx.tinit.has_bounds:
            fh.status.append('004')
        # Check consistency between averaged time and time boundaries
        if not ctx.tinit.is_instant and not ctx.tinit.has_bounds:
            fh.status.append('005')
        # Check time axis squareness
        if not np.array_equal(fh.time_axis_rebuilt, fh.time_axis):
            fh.status.append('001')
        else:
            fh.status.append('000')
        # Check time boundaries squareness if needed
        if ctx.tinit.has_bounds:
            fh.time_bounds_rebuilt = fh.build_time_bounds(start=trunc(start, 5),
                                                          inc=time_inc(ctx.tinit.frequency)[0],
                                                          input_units=ctx.tinit.funits,
                                                          output_units=ctx.tinit.tunits,
                                                          calendar=ctx.tinit.calendar)
            if not np.array_equal(fh.time_bounds_rebuilt, fh.time_bounds):
                fh.status.append('006')
        # Check consistency between time units
        if ctx.tinit.tunits != fh.time_units:
            fh.status.append('002')
        # Check consistency between time units
        if ctx.tinit.calendar != fh.calendar:
            fh.status.append('007')
        # Rename file depending on checking
        if (ctx.write or ctx.force) and {'003'}.intersection(set(fh.status)):
            # Change filename and file full path dynamically
            fh.nc_file_rename(new_filename=re.sub(fh.end_timestamp, fh.last_timestamp, fh.filename))
        # Remove time boundaries depending on checking
        if (ctx.write or ctx.force) and {'004'}.intersection(set(fh.status)):
            # Delete time bounds and bounds attribute from file if write of force mode
            fh.nc_var_delete(variable=TIME_BNDS_VAR_NAME[project])
            fh.nc_att_delete(attribute='bounds', variable='time')
        # Rewrite time axis depending on checking
        if (ctx.write and {'001', '002', '006', '007'}.intersection(set(fh.status))) or ctx.force:
            fh.nc_var_overwrite('time', fh.time_axis_rebuilt)
            fh.nc_att_overwrite('units', 'time', ctx.tinit.tunits)
            fh.nc_att_overwrite('calendar', 'time', ctx.tinit.calendar)
            # Rewrite time boundaries if needed
            if ctx.tinit.has_bounds:
                fh.nc_var_overwrite(TIME_BNDS_VAR_NAME[project], fh.time_bounds_rebuilt)
        # Compute checksum at the end of all modifications and after closing file
        if (ctx.write or ctx.force) and {'001', '002', '003', '004', '006', '007'}.intersection(set(fh.status)):
            fh.new_checksum = fh.checksum(ctx.checksum_type)
        # Print file status
        msg = """Filename: {}
                                   Start: {} = {}
                                   End:   {} = {}
                                   Last:  {} = {}
                                   Time steps: {}
                                   Is instant: {}""".format(fh.filename,
                                                            fh.start_timestamp, fh.start_date,
                                                            fh.end_timestamp, fh.end_date,
                                                            fh.last_timestamp, fh.last_date,
                                                            fh.length,
                                                            ctx.tinit.is_instant)
        for s in fh.status:
            msg += """\n                                   Status: {}""".format(STATUS[s])
        if not {'000', '008'}.intersection(set(fh.status)):
            logging.error(msg)
        else:
            if {'008'}.intersection(set(fh.status)):
                logging.warning(msg)
            else:
                logging.info(msg)
        # Return file status
        return fh
    except Exception as e:
        ctx.status.append('999')
        logging.error('{} skipped\n{}: {}'.format(ffp, e.__class__.__name__, e.message))
        return None


def run(args):
    """
    Main process that:

     * Instantiates processing context,
     * Defines the referenced time properties,
     * Instantiates threads pools,
     * Prints or logs the time axis diagnostics.

    :param ArgumentParser args: Command-line arguments parser

    """
    for directory in args.directory:
        # Instantiate processing context
        with ProcessingContext(args, directory) as ctx:
            logging.info('Time axis diagnostic started for {}'.format(ctx.directory))
            # Process supplied files
            handlers = [x for x in ctx.pool.imap(process, ctx.sources)]
            ctx.scan_files = len(handlers)
            # Persist diagnostics into database
            if ctx.db:
                # Check if database exists
                if not os.path.isfile(ctx.db):
                    logging.warning('Database does not exist')
                    db.create(ctx.db)
            # Commit each diagnostic as a new entry
            for handler in handlers:
                ctx.status.extend(handler.status)
                if ctx.db:
                    diagnostic = dict()
                    diagnostic['creation_date'] = datetime(1, 1, 1)._to_real_datetime().now().strftime("%Y%m%d-%H%M%S")
                    diagnostic.update(ctx.__dict__)
                    diagnostic.update(ctx.tinit.__dict__)
                    diagnostic.update(handler.__dict__)
                    diagnostic['status'] = ','.join(handler.status)
                    db.insert(ctx.db, diagnostic)
                    logging.info('{} - Diagnostic persisted into database'.format(handler.filename))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Constants used in this module.

"""

# Database connexion timeout (in sec)
TIMEOUT = 120

# Default maximum threads number
MAX_THREADS_DEFAULT = 1

# Errors that return exit code = 1
EXIT_ERRORS = ['005','999']

# Status messages
STATUS = {'000': 'Time axis seems OK',
          '001': 'Mistaken time axis over one or several time steps',
          '002': 'Time units must be unchanged for the same dataset',
          '003': 'Last timestamp differs from end timestamp of filename',
          '004': 'An instantaneous time axis should not embed time boundaries',
          '005': 'An averaged time axis should embed time boundaries',
          '006': 'Mistaken time bounds over one or several time steps',
          '007': 'Calendar must be unchanged for the same dataset',
          '008': 'Last date differs from end date of filename',
          '999': 'Other error'}




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Processing context used in this module.

"""
import logging
import os
import sys

import networkx as nx
from ESGConfigParser import SectionParser

from nctime.utils.collector import Collector
from nctime.utils.constants import *
from nctime.utils.custom_exceptions import *
from nctime.utils.misc import cmd_exists
from nctime.utils.time import TimeInit


class ProcessingContext(object):
    """
    Encapsulates the processing context/information for main process.

    :param ArgumentParser args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *ProcessingContext*

    """

    def __init__(self, args, directory):
        self.directory = directory
        self.config_dir = args.i
        self.resolve = args.resolve
        self.full_overlap_only = args.full_overlap_only
        self.verbose = args.v
        self.project = args.project
        self.tunits_default = None
        if self.project in DEFAULT_TIME_UNITS.keys():
            self.tunits_default = DEFAULT_TIME_UNITS[self.project]
        self.path = []
        self.full_overlaps = None
        self.partial_overlaps = None
        self.broken = False

    def __enter__(self):
        # Get checksum client
        self.checksum_client, self.checksum_type = self.get_checksum_client()
        # Init configuration parser
        self.cfg = SectionParser(section='project:{}'.format(self.project), directory=self.config_dir)
        self.pattern = self.cfg.translate('filename_format')
        # Init data collector
        self.sources = Collector(source=self.directory)
        # Init collector filter
        # Exclude hidden non-NetCDF files
        self.sources.FileFilter['base_filter'] = ('^[!.].*\.nc$', True)
        # Exclude fixed frequency
        self.sources.FileFilter['frequency_filter'] = ('(_fx_|_fixed_|_fx.|_fixed.)', True)
        # Get first file for reference
        self.ref = self.sources.first()
        self.display = len(os.path.basename(self.ref))
        # Set driving time properties
        self.tinit = TimeInit(project=self.project, ref=self.ref, tunits_default=self.tunits_default)
        # DiGraph creation
        self.graph = nx.DiGraph()
        return self

    def __exit__(self, *exc):
        # Decline outputs depending on the scan results
        # Default is sys.exit(0)
        # Print first node
        m = ' START '.center(self.display + 2, '-')
        msg = '\n                                   {}'.format(m)
        # Print intermediate nodes
        for i in range(1, len(self.path) - 2):
            m = ' {} '.format(self.path[i]).center(self.display + 2, '~')
            if self.partial_overlaps and self.path[i] in self.partial_overlaps:
                m = ' {} < overlap from {} to {} '.format(self.path[i].center(self.display + 2),
                                                          self.partial_overlaps[self.path[i]]['start_date'],
                                                          self.partial_overlaps[self.path[i]]['end_overlap'])
            msg += '\n                                   {}'.format(m)
        # Print last node
        m = ' END '.center(self.display + 2, '-')
        msg += '\n                                   {}'.format(m)
        # Print analyse result
        if self.broken:
            logging.error('Time series broken: {}'.format(msg))
            sys.exit(1)
        else:
            logging.info('Shortest path found: {}'.format(msg))
        # Print overlaps if exists
        if not self.full_overlaps and not self.partial_overlaps:
            logging.info('No overlapping files')
        else:
            logging.warning('Overlapping files:')
            for node in self.full_overlaps:
                if self.resolve:
                    logging.warning('{} > REMOVED'.format(node))
                else:
                    logging.warning('{} > TO REMOVE'.format(node))
            for node in self.partial_overlaps:
                if self.resolve and not self.full_overlap_only:
                    logging.warning('{} > TRUNCATED'.format(node))
                else:
                    logging.warning('{} > TO TRUNCATE'.format(node))
        # End
        logging.info('Overlap diagnostic completed')

    def get_checksum_client(self):
        """
        Gets the checksum client to use.
        Be careful to Exception constants by reading two different sections.

        :returns: The checksum client
        :rtype: *str*

        """
        _cfg = SectionParser(section='DEFAULT', directory=self.config_dir)
        if _cfg.has_option('DEFAULT', 'checksum'):
            checksum_client, checksum_type = _cfg.get_options_from_table('checksum')[0]
        else:  # Use SHA256 as default because esg.ini not mandatory in configuration directory
            checksum_client, checksum_type = 'sha256sum', 'SHA256'
        if not cmd_exists(checksum_client):
            raise ChecksumClientNotFound(checksum_client)
        else:
            return checksum_client, checksum_type



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
   :platform: Unix
   :synopsis: Highlight chunked NetCDF files producing overlap in a time series.

"""

import logging
import os
import re

import nco
import networkx as nx
import numpy as np

from context import ProcessingContext
from nctime.utils.time import get_next_timestep, get_start_end_dates_from_filename, \
    get_first_last_timesteps, dates2int
from itertools import combinations, permutations

def get_overlaps(directory, nodes, shortest):
    """
    Returns all overlapping (default) files as a list of tuples. Each tuple gathers\:
     * The higher overlap bound,
     * The date to cut the file in order to resolve the overlap,
     * The corresponding cutting timestep.

    :param str directory: The directory scanned
    :param dict nodes: The directed graph nodes as a dictionary
    :param list shortest: The most consecutive files list (from `nx.DiGraph()`)
    :returns: The filenames
    :rtype: *list*

    """
    overlaps = dict()
    # Find partial overlapping nodes
    overlaps['partial'] = dict()
    for n in range(1, len(shortest) - 2):
        # Get current node dates
        current_node = nodes[shortest[n]]
        # Get next node dates
        next_node = nodes[shortest[n + 1]]
        # Because of the shortest path is selected, difference between bounds should always be positive or zero
        assert (current_node['next_date'] - next_node['start_date']) >= 0
        # Get partial overlap if exists
        # Partial overlap from next_node[1] to current_node[2] (bounds included)
        # Overlap is hold on the next node (arbitrary)
        if (current_node['next_date'] - next_node['start_date']) > 0:
            cutting_timestep = get_next_timestep(os.path.join(directory, shortest[n + 1]),
                                                 current_node['last_timestep'])
            overlaps['partial'][shortest[n + 1]] = next_node
            overlaps['partial'][shortest[n + 1]].update({'end_overlap': current_node['end_date'],
                                                         'cutting_date': current_node['next_date'],
                                                         'cutting_timestep': cutting_timestep})
    # Find full overlapping nodes
    overlaps['full'] = list(set(nodes.keys()).symmetric_difference(set(shortest[1:-1])))
    return overlaps['full'], overlaps['partial']


def resolve_overlap(directory, pattern, filename, from_date=None, to_date=None, cutting_timestep=None, partial=False):
    """
    Resolve overlapping files.
    If full overlap, the corresponding file is removed.
    If partial overlap, the corresponding file is truncated into a new one and the old file is removed.

    :param str directory: The directory scanned
    :param str pattern: The filename pattern
    :param str filename: The filename
    :param int from_date: Overlap starting date
    :param int to_date:  Overlap ending date
    :param int cutting_timestep:  Time step to cut the file
    :param boolean partial: Resolve partial overlap if True

    """
    if partial:
        filename_attr = re.match(pattern, filename).groupdict()
        assert len(filename_attr['period_start']) == len(filename_attr['period_end'])
        from_timestamp = str(from_date)[:len(filename_attr['period_start'])]
        to_timestamp = str(to_date)[:len(filename_attr['period_end'])]
        tmp = filename.replace(filename_attr['period_start'], from_timestamp)
        new_filename = tmp.replace(filename_attr['period_end'], to_timestamp)
        assert not os.path.exists(os.path.join(directory, new_filename))
        nc = nco.Nco()
        nc.ncks(input=os.path.join(directory, filename),
                output=os.path.join(directory, new_filename),
                options='-O -d time,{},,1'.format(cutting_timestep))
    os.remove(os.path.join(directory, filename))


def run(args):
    """
    Main process that:

     * Instantiates processing context,
     * Deduces start, end and next date from each filenames,
     * Builds the DiGraph,
     * Detects the shortest path between dates if exists,
     * Detects broken path between dates if exists,
     * Removes the overlapping files.

    :param ArgumentParser args: Command-line arguments parser

    """
    for directory in args.directory:
        # Instantiate processing context
        with ProcessingContext(args, directory) as ctx:
            logging.info('Overlap diagnostic started for {}'.format(ctx.directory))
            ctx.graph.clear()
            nodes = dict()
            logging.info('Create edges from source nodes to target nodes')
            for ffp in ctx.sources:
                filename = os.path.basename(ffp)
                # A node is defined by the filename, its start, end and next dates with its first and last timesteps.
                nodes[filename] = dict()
                # Get start, end and next date of each file
                dates = get_start_end_dates_from_filename(filename=filename,
                                                          pattern=ctx.pattern,
                                                          frequency=ctx.tinit.frequency,
                                                          calendar=ctx.tinit.calendar)
                nodes[filename]['start_date'], nodes[filename]['end_date'], nodes[filename]['next_date'] = dates2int(
                    dates)
                # Get first and last time steps
                timesteps = get_first_last_timesteps(ffp)
                nodes[filename]['first_timestep'], nodes[filename]['last_timestep'] = timesteps

            # Build starting node
            start_dates = [nodes[n]['start_date'] for n in nodes]
            starts = [n for n in nodes if nodes[n]['start_date'] == min(start_dates)]
            for start in starts:
                ctx.graph.add_edge('START', start)
                if ctx.verbose:
                    logging.info('{} --> {}'.format(' START '.center(ctx.display + 2, '-'),
                                                    start.center(ctx.display + 2)))
            # Create graph edges with filenames only
            end_dates = [nodes[n]['end_date'] for n in nodes]
            other_nodes = []
            for n in nodes:
                # Create a fake ending node from all latest files in case of overlaps
                if nodes[n]['end_date'] == max(end_dates):
                    ctx.graph.add_edge(n, 'END')
                    if ctx.verbose:
                        logging.info('{} --> {}'.format(n.center(ctx.display + 2),
                                                        ' END '.center(ctx.display + 2, '-')))
                else:
                    # A node is a "backward" node when the difference between the next current time step
                    # and its start date is positive.
                    # To ensure continuity path, edges has to only exist with backward nodes.
                    other_nodes = [x for x in nodes if x is not n]
                    start_dates = [nodes[x]['start_date'] for x in other_nodes]
                    # Find index of "backwards" nodes
                    nxts = [i for i, v in enumerate(nodes[n]['next_date'] - np.array(start_dates)) if v >= 0]
                    nxts = [other_nodes[i] for i in nxts]
                    for nxt in nxts:
                        ctx.graph.add_edge(n, nxt)
                        if ctx.verbose:
                            logging.info('{} --> {}'.format(n.center(ctx.display + 2),
                                                            nxt.center(ctx.display + 2)))
            # Walk through the graph
            try:
                # Find shortest path between oldest and latest dates
                ctx.path = nx.shortest_path(ctx.graph, source='START', target='END')
                # Get overlaps
                ctx.full_overlaps, ctx.partial_overlaps = get_overlaps(ctx.directory, nodes, ctx.path)
            except nx.NetworkXNoPath:
                ctx.broken = True
                ctx.path.append('START')
                for n in sorted(nodes.keys()):
                    ctx.path.append(n)
                    # A node is a "forward" node when the difference between the next current time step
                    # and its start date is negative.
                    # A path gap exists from a node when no edges exist with ALL "forwards" nodes.
                    other_nodes = [x for x in nodes if x is not n]
                    start_dates = [nodes[x]['start_date'] for x in other_nodes]
                    # Find index of "forward" nodes
                    nxts = [i for i, v in enumerate(nodes[n]['next_date'] - np.array(start_dates)) if v <= 0]
                    nxts = [other_nodes[i] for i in nxts]
                    # Find available targets from graph
                    try:
                        avail_targets = zip(*ctx.graph.edges(n))[1]
                    except IndexError:
                        avail_targets = []
                    # If no "forward" nodes in edges target = BREAK
                    if not set(nxts).intersection(avail_targets):
                        ctx.path.append('BREAK')
                ctx.path.append('END')
            # Resolve overlaps
            if ctx.resolve:
                # Full overlapping files has to be deleted before partial overlapping files are truncated.
                for node in ctx.full_overlaps:
                    resolve_overlap(directory=ctx.directory,
                                    pattern=ctx.pattern,
                                    filename=node,
                                    partial=False)
                if not ctx.full_overlap_only:
                    for node in ctx.partial_overlaps:
                        resolve_overlap(directory=ctx.directory,
                                        pattern=ctx.pattern,
                                        filename=node,
                                        from_date=ctx.overlaps['partial'][node]['cutting_date'],
                                        to_date=ctx.overlaps['partial'][node]['end_date'],
                                        cutting_timestep=ctx.overlaps['partial'][node]['cutting_timestep'],
                                        partial=True)



#!/usr/bin/env python
"""
.. module:: nctime.axis.handler.py
   :platform: Unix
   :synopsis: File handler for time axis.

"""

import os
import re
from uuid import uuid4

import nco
import netCDF4
import numpy as np

from custom_exceptions import *
from nctime.utils.custom_exceptions import *
from nctime.utils.time import truncated_timestamp, get_start_end_dates_from_filename, dates2str, num2date, date2num


class File(object):
    """
    Handler providing methods to deal with file processing.

    :returns: The axis status
    :rtype: *File*

    """

    def __init__(self, ffp, has_bounds):
        self.ffp = ffp
        # Retrieve directory and filename full path
        self.directory, self.filename = os.path.split(ffp)
        # Start/end period dates from filename + next/last expected dates
        self.start_date = None
        self.end_date = None
        self.next_date = None
        self.last_date = None
        # Filename timestamp length
        self.timestamp_length = None
        # Start/end period timestamp from filename + next/last expected timestamps
        self.start_timestamp = None
        self.end_timestamp = None
        self.next_timestamp = None
        self.last_timestamp = None
        # Get time axis length
        try:
            f = netCDF4.Dataset(self.ffp, 'r')
        except IOError:
            raise InvalidNetCDFFile(self.ffp)
        self.length = f.variables['time'].shape[0]
        # Get time boundaries
        if has_bounds:
            try:
                self.time_bounds = f.variables['time_bnds'][:, :]
            except KeyError:
                self.time_bounds = f.variables['time_bounds'][:, :]
        else:
            self.time_bounds = None
        # Get time axis
        self.time_axis = f.variables['time'][:]
        # Get time units from file
        self.time_units = f.variables['time'].units
        # Get calendar from file
        self.calendar = f.variables['time'].calendar
        f.close()
        self.time_axis_rebuilt = None
        self.time_bounds_rebuilt = None
        self.status = list()
        self.new_checksum = None

    def get_start_end_dates(self, pattern, frequency, units, calendar):
        """
        Wraps and records :func:`get_start_end_dates_from_filename` results.

        :param re Object pattern: The filename pattern as a regex (from `re library \
        <https://docs.python.org/2/library/re.html>`_).
        :param str frequency: The time frequency
        :param str units: The proper time units
        :param str calendar: The NetCDF calendar attribute
        :returns: Start and end dates as number of days since the referenced date
        :rtype: *float*

        """
        self.timestamp_length = len(re.match(pattern, self.filename).groupdict()['period_end'])
        dates = get_start_end_dates_from_filename(filename=self.filename,
                                                  pattern=pattern,
                                                  frequency=frequency,
                                                  calendar=calendar)
        self.start_date, self.end_date, self.next_date = dates2str(dates)
        self.start_timestamp, self.end_timestamp, self.next_timestamp = [
            truncated_timestamp(date, self.timestamp_length) for date in dates]
        return date2num(dates, units=units, calendar=calendar)

    def checksum(self, checksum_type):
        """
        Does the checksum by the Shell avoiding Python memory limits.

        :param str checksum_type: Checksum type
        :returns: The checksum
        :rtype: *str*
        :raises Error: If the checksum fails

        """
        checksum_client = {'SHA256': 'sha256sum',
                           'MD5': 'md5sum'}
        assert (checksum_type in checksum_client.keys()), 'Invalid checksum type'
        try:
            shell = os.popen("{} {} | awk -F ' ' '{{ print $1 }}'".format(checksum_client[checksum_type],
                                                                          self.ffp),
                             'r')
            return shell.readline()[:-1]
        except:
            raise ChecksumFail(checksum_type, self.ffp)

    def build_time_axis(self, start, inc, input_units, output_units, calendar, is_instant=False):
        """
        Rebuilds time axis from date axis, depending on MIP frequency, calendar and instant status.

        :param float start: The numerical date to start (from ``netCDF4.date2num`` or \
        :func:`nctime.utils.time.date2num`)
        :param int inc: The time incrementation
        :param input_units: The time units deduced from the frequency
        :param output_units: The time units from the file
        :param calendar: The time calendar fro NetCDF attributes
        :param boolean is_instant: True if instantaneous time axis
        :returns: The corresponding theoretical time axis
        :rtype: *numpy.array*

        """
        num_axis = np.arange(start=start, stop=start + self.length * inc, step=inc)
        if input_units.split(' ')[0] in ['years', 'months']:
            last_date = num2date(num_axis[-1], units=input_units, calendar=calendar)[0]
        else:
            last_date = num2date(num_axis[-1], units=input_units, calendar=calendar)
        self.last_date = dates2str(last_date)
        self.last_timestamp = truncated_timestamp(last_date, self.timestamp_length)
        if not is_instant:
            num_axis += 0.5 * inc
        date_axis = num2date(num_axis, units=input_units, calendar=calendar)
        return date2num(date_axis, units=output_units, calendar=calendar)

    def build_time_bounds(self, start, inc, input_units, output_units, calendar):
        """
        Rebuilds time boundaries from the start date, depending on MIP frequency, calendar and
        instant status.

        :param float start: The numerical date to start (from ``netCDF4.date2num`` or \
        :func:`nctime.utils.time.date2num`)
        :param int inc: The time incrementation
        :param input_units: The time units deduced from the frequency
        :param output_units: The time units from the file
        :param calendar: The time calendar fro NetCDF attributes

        :returns: The corresponding theoretical time boundaries as a [n, 2] array
        :rtype: *numpy.array*

        """
        num_axis_bnds = np.column_stack(((np.arange(start=start,
                                                    stop=start + self.length * inc,
                                                    step=inc)),
                                         (np.arange(start=start,
                                                    stop=start + (self.length + 1) * inc,
                                                    step=inc)[1:])))
        date_axis_bnds = num2date(num_axis_bnds, units=input_units, calendar=calendar)
        return date2num(date_axis_bnds, units=output_units, calendar=calendar)

    def nc_var_delete(self, variable):
        """
        Delete a NetCDF variable using NCO operators.
        A unique filename is generated to avoid multithreading errors.
        To overwrite the input file, the source file is dump using the ``cat`` Shell command-line
        to avoid Python memory limit.

        :param str variable: The variable to delete
        :raises Error: If the deletion failed

        """
        # Generate unique filename
        fftmp = '{}/{}{}'.format(self.directory, str(uuid4()), '.nc')
        try:
            nc = nco.Nco()
            nc.ncks(input=self.ffp,
                    output=fftmp,
                    options='-x -O -v {}'.format(variable))
            os.popen("cat {} > {}".format(fftmp, self.ffp), 'r')
            os.remove(fftmp)
        except:
            os.remove(fftmp)
            raise NetCDFVariableRemoveFail(variable, self.ffp)

    def nc_att_delete(self, variable, attribute):
        """
        Delete a NetCDF dimension attribute using NCO operators.
        A unique filename is generated to avoid multithreading errors.
        To overwrite the input file, the source file is dump using the ``cat`` Shell command-line
        to avoid Python memory limit.

        :param str attribute: The attribute to delete
        :param str variable: The variable that has the attribute
        :raises Error: If the deletion failed

        """
        # Generate unique filename
        fftmp = '{}/{}{}'.format(self.directory, str(uuid4()), '.nc')
        try:
            nc = nco.Nco()
            nc.ncatted(input=self.ffp,
                       output=fftmp,
                       options='-a {},{},d,,'.format(attribute, variable))
            os.popen("cat {} > {}".format(fftmp, self.ffp), 'r')
            os.remove(fftmp)
        except:
            os.remove(fftmp)
            raise NetCDFAttributeRemoveFail(attribute, self.ffp, variable)

    def nc_var_overwrite(self, variable, data):
        """
        Rewrite variable to NetCDF file without copy.

        :param str variable: The variable to replace
        :param float array data: The data array to overwrite

        """
        try:
            f = netCDF4.Dataset(self.ffp, 'r+')
        except IOError:
            raise InvalidNetCDFFile(self.ffp)
        f.variables[variable][:] = data
        f.close()

    def nc_att_overwrite(self, attribute, variable, data):
        """
        Rewrite attribute to NetCDF file without copy.


        :param str attribute: THe attribute to replace
        :param str variable: The variable that has the attribute
        :param str data: The string to add to overwrite

        """
        try:
            f = netCDF4.Dataset(self.ffp, 'r+')
        except IOError:
            raise InvalidNetCDFFile(self.ffp)
        setattr(f.variables[variable], attribute, data)
        f.close()

    def nc_file_rename(self, new_filename):
        """
        Rename a NetCDF file.

        :param str new_filename: The new filename to apply

        """
        ffp = os.path.join(os.path.dirname(self.ffp), new_filename)
        if os.path.exists(ffp):
            raise RenamingNetCDFFailed(self.ffp, ffp, exists=True)
        try:
            os.rename(self.ffp, ffp)
        except OSError:
            raise RenamingNetCDFFailed(self.ffp, ffp)
        self.ffp = ffp
        self.filename = new_filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Methods used to deal with NetCDF file time axis.

"""

import logging
import os
import re

import netCDF4
import numpy as np
from netcdftime import utime

from constants import *
from custom_exceptions import *


class TimeInit(object):
    """
    Encapsulates the time properties from first file into processing context.
    These properties has to be used as reference for all files into the directory.

     * The calendar, the frequency and the realm are read from NetCDF global attributes and \
     use to detect instantaneous time axis,
     * The NetCDF time units attribute has to be unchanged in respect with CF convention and \
     archives designs.

    :param str ref: The reference file full path
    :param str tunits_default: The default time units if exists
    :raises Error: If NetCDF time units attribute is missing
    :raises Error: If NetCDF frequency attribute is missing
    :raises Error: If NetCDF realm attribute is missing
    :raises Error: If NetCDF calendar attribute is missing

    """

    def __init__(self, project, ref, tunits_default=None):
        variable = unicode(os.path.basename(ref).split('_')[0])
        try:
            nc = netCDF4.Dataset(ref, 'r')
        except IOError:
            raise InvalidNetCDFFile(ref)
        # Check and set required global attributes
        for attr, name in REQUIRED_ATTRIBUTES[project].iteritems():
            if name not in nc.ncattrs():
                raise NoNetCDFAttribute(name, ref)
            else:
                setattr(self, attr, getattr(nc, name))
        # Check time variable exists
        if 'time' not in nc.variables.keys():
            raise NoNetCDFVariable('time', ref)
        # Check required time attributes exist
        for attribute in REQUIRED_TIME_ATTRIBUTES:
            if attribute not in nc.variables['time'].ncattrs():
                raise NoNetCDFAttribute(attribute, ref, 'time')
        # Get time units (i.e., days since ...)
        self.tunits = self.control_time_units(nc.variables['time'].units, tunits_default)
        # Convert time units into frequency units (i.e., months/year/hours since ...)
        self.funits = self.convert_time_units(self.tunits, self.frequency)
        # Get calendar
        self.calendar = nc.variables['time'].calendar
        # Get boolean on instantaneous time axis
        if 'cell_methods' not in nc.variables[variable].ncattrs():
            raise NoNetCDFAttribute('cell_methods', ref, variable)
        self.is_instant = False
        if 'point' in nc.variables[variable].cell_methods.lower():
            self.is_instant = True
        # Get boolean on time boundaries
        self.has_bounds = False
        if 'time_bnds' or 'time_bounds' in nc.variables.keys():
            self.has_bounds = True
        nc.close()

    @staticmethod
    def control_time_units(tunits, tunits_default=None):
        """
        Controls the time units format as at least "days since YYYY-MM-DD".
        The time units can be forced within configuration file using the ``time_units_default`` option.

        :param str tunits: The NetCDF time units string from file
        :param tunits_default: The default time units that should be used
        :returns: The appropriate time units string formatted and controlled depending on the project
        :rtype: *str*

        """
        units = tunits.split()
        units[0] = unicode('days')
        if len(units[2].split('-')) == 1:
            units[2] += '-{}-{}'.format('01', '01')
        elif len(units[2].split('-')) == 2:
            units[2] += '-{}'.format('01')
        if tunits_default and ' '.join(units) != tunits_default:
            logging.warning('Invalid time units. Replace "{}" by "{}"'.format(' '.join(units), tunits_default))
            return tunits_default
        else:
            return ' '.join(units)

    @staticmethod
    def convert_time_units(tunits, frequency):
        """
        Converts default time units from file into time units using the MIP frequency.
        As en example, for a 3-hourly file, the time units "days since YYYY-MM-DD" becomes
        "hours since YYYY-MM-DD".

        :param str tunits: The NetCDF time units string from file
        :param str frequency: The time frequency
        :returns: The converted time units string
        :rtype: *str*

        """
        units = {'subhr': 'minutes',
                 '3hr': 'hours',
                 '6hr': 'hours',
                 'day': 'days',
                 'mon': 'months',
                 'yr': 'years'}
        return tunits.replace('days', units[frequency])


def untruncated_timestamp(timestamp):
    """
    Returns proper digits for yearly and monthly truncated timestamps.
    The dates from filename are filled with the 0 digit to reach 14 digits.
    Consequently, yearly dates starts at January 1st and monthly dates starts at first day
    of the month.

    :param str timestamp: A date string from a filename
    :returns: The filled timestamp
    :rtype: *str*

    """
    if len(timestamp) == 4:
        # Start year at january 1st
        return (timestamp + '0101').ljust(14, '0')
    elif len(timestamp) == 6:
        # Start month at first day
        return (timestamp + '01').ljust(14, '0')
    else:
        return timestamp.ljust(14, '0')


def truncated_timestamp(date, length):
    """
    Returns proper digits depending on datetime object.

    :param datetime.datetime date: Datetime or phony datetime object
    :param int length: The timestamp length expected
    :returns: The corresponding timestamp
    :rtype: *str*

    """
    timestamp = date2str(date, iso_format=False)
    return timestamp[:length]


def num2date(num_axis, units, calendar):
    """
    A wrapper from ``netCDF4.num2date`` able to handle "years since" and "months since" units.
    If time units are not "years since" or "months since", calls usual ``netcdftime.num2date``.

    :param numpy.array num_axis: The numerical time axis following units
    :param str units: The proper time units
    :param str calendar: The NetCDF calendar attribute
    :returns: The corresponding date axis
    :rtype: *array*

    """
    if not units.split(' ')[0] in ['years', 'months']:
        # If units are not 'years' or 'months since', call usual netcdftime.num2date:
        return netCDF4.num2date(num_axis, units=units, calendar=calendar)
    else:
        # Return to time reference with 'days since'
        units_as_days = 'days ' + ' '.join(units.split(' ')[1:])
        # Convert the time reference 'units_as_days' as datetime object
        start_date = netCDF4.num2date(0.0, units=units_as_days, calendar=calendar)
        # Control num_axis to always get an Numpy array (even with a scalar)
        num_axis_mod = np.atleast_1d(np.array(num_axis))
        if units.split(' ')[0] == 'years':
            # If units are 'years since'
            # Define the number of maximum and minimum years to build a date axis covering
            # the whole 'num_axis' period
            max_years = np.floor(np.max(num_axis_mod)) + 1
            min_years = np.ceil(np.min(num_axis_mod)) - 1
            # Create a date axis with one year that spans the entire period by year
            years_axis = np.array([add_year(start_date, years_to_add)
                                   for years_to_add in np.arange(min_years, max_years + 2)])
            # Convert rebuilt years axis as 'number of days since'
            cdftime = utime(units_as_days, calendar=calendar)
            years_axis_as_days = cdftime.date2num(years_axis)
            # Index of each years
            yind = np.vectorize(np.int)(np.floor(num_axis_mod))
            # Rebuilt num_axis as 'days since' adding the number of days since referenced time
            # with an half-increment (num_axis_mod - yind) = 0 or 0.5
            num_axis_mod_days = (years_axis_as_days[yind - int(min_years)] +
                                 (num_axis_mod - yind) *
                                 np.diff(years_axis_as_days)[yind - int(min_years)])
            # Convert result as date axis
            return netCDF4.num2date(num_axis_mod_days, units=units_as_days, calendar=calendar)
        elif units.split(' ')[0] == 'months':
            # If units are 'months since'
            # Define the number of maximum and minimum months to build a date axis covering
            # the whole 'num_axis' period
            max_months = np.floor(np.max(num_axis_mod)) + 1
            min_months = np.ceil(np.min(num_axis_mod)) - 1
            # Create a date axis with one month that spans the entire period by month
            months_axis = np.array([add_month(start_date, months_to_add)
                                    for months_to_add in np.arange(min_months, max_months + 12)])
            # Convert rebuilt months axis as 'number of days since'
            cdftime = utime(units_as_days, calendar=calendar)
            months_axis_as_days = cdftime.date2num(months_axis)
            # Index of each months
            mind = np.vectorize(np.int)(np.floor(num_axis_mod))
            # Rebuilt num_axis as 'days since' adding the number of days since referenced time
            # with an half-increment (num_axis_mod - mind) = 0 or 0.5
            num_axis_mod_days = (months_axis_as_days[mind - int(min_months)] +
                                 (num_axis_mod - mind) *
                                 np.diff(months_axis_as_days)[mind - int(min_months)])
            # Convert result as date axis
            return netCDF4.num2date(num_axis_mod_days, units=units_as_days, calendar=calendar)


def date2num(date_axis, units, calendar):
    """
    A wrapper from ``netCDF4.date2num`` able to handle "years since" and "months since" units.
    If time units are not "years since" or "months since" calls usual ``netcdftime.date2num``.

    :param numpy.array date_axis: The date axis following units
    :param str units: The proper time units
    :param str calendar: The NetCDF calendar attribute
    :returns: The corresponding numerical time axis
    :rtype: *array*

    """
    # date_axis is the date time axis incremented following units (i.e., by years, months, etc).
    if not units.split(' ')[0] in ['years', 'months']:
        # If units are not 'years' or 'months since', call usual netcdftime.date2num:
        return netCDF4.date2num(date_axis, units=units, calendar=calendar)
    else:
        # Return to time reference with 'days since'
        units_as_days = 'days ' + ' '.join(units.split(' ')[1:])
        # Convert date axis as number of days since time reference
        days_axis = netCDF4.date2num(date_axis, units=units_as_days, calendar=calendar)
        # Convert the time reference 'units_as_days' as datetime object
        start_date = netCDF4.num2date(0.0, units=units_as_days, calendar=calendar)
        # Create years axis from input date axis
        years = np.array([date.year for date in np.atleast_1d(np.array(date_axis))])
        if units.split(' ')[0] == 'years':
            # If units are 'years since'
            # Define the number of maximum and minimum years to build a date axis covering
            # the whole 'num_axis' period
            max_years = np.max(years - start_date.year + 1)
            min_years = np.min(years - start_date.year - 1)
            # Create a date axis with one year that spans the entire period by year
            years_axis = np.array([add_year(start_date, yid)
                                   for yid in np.arange(min_years, max_years + 2)])
            # Convert years axis as number of days since time reference
            cdftime = utime(units_as_days, calendar=calendar)
            years_axis_as_days = cdftime.date2num(years_axis)
            # Find closest index for years_axis_as_days in days_axis
            closest_index = np.searchsorted(years_axis_as_days, days_axis)
            # Compute the difference between closest value of year axis and start date, in number of days
            num = days_axis - years_axis_as_days[closest_index]
            # Number of days of the corresponding closest year
            den = np.diff(years_axis_as_days)[closest_index]
            return min_years + closest_index + num / den
        elif units.split(' ')[0] == 'months':
            # If units are 'months since'
            # Define the number of maximum and minimum months to build a date axis covering
            # the whole 'num_axis' period
            max_months = np.max(12 * (years - start_date.year + 1))
            min_months = np.min(12 * (years - start_date.year - 1))
            # Create a date axis with one month that spans the entire period by month
            months_axis = np.array([add_month(start_date, mid)
                                    for mid in np.arange(min_months, max_months)])
            # Convert months axis as number of days since time reference
            cdftime = utime(units_as_days, calendar=calendar)
            months_axis_as_days = cdftime.date2num(months_axis)
            # Find closest index for months_axis_as_days in days_axis
            closest_index = np.searchsorted(months_axis_as_days, days_axis)
            # Compute the difference between closest value of months axis and start date, in number of days
            num = days_axis - months_axis_as_days[closest_index]
            # Number of days of the corresponding closest month
            den = np.diff(months_axis_as_days)[closest_index]
            return min_months + closest_index + num / den


def add_month(date, months_to_add):
    """
    Finds the next month from date.

    :param netcdftime.datetime date: Accepts datetime or phony datetime from ``netCDF4.num2date``.
    :param int months_to_add: The number of months to add to the date
    :returns: The final date
    :rtype: *netcdftime.datetime*

    """
    years_to_add = int((date.month + months_to_add - np.mod(date.month + months_to_add - 1, 12) - 1) / 12)
    new_month = int(np.mod(date.month + months_to_add - 1, 12)) + 1
    new_year = date.year + years_to_add
    date_next = datetime(year=new_year,
                         month=new_month,
                         day=date.day,
                         hour=date.hour,
                         minute=date.minute,
                         second=date.second)

    return date_next


def add_year(date, years_to_add):
    """
    Finds the next year from date.

    :param netcdftime.datetime date: Accepts datetime or phony datetime from ``netCDF4.num2date``.
    :param int years_to_add: The number of years to add to the date
    :returns: The final date
    :rtype: *netcdftime.datetime*

    """
    new_year = date.year + years_to_add
    date_next = datetime(year=new_year,
                         month=date.month,
                         day=date.day,
                         hour=date.hour,
                         minute=date.minute,
                         second=date.second)
    return date_next


def get_start_end_dates_from_filename(filename, pattern, frequency, calendar):
    """
    Returns datetime objects for start and end dates from the filename.
    To rebuild a proper time axis, the dates from filename are expected to set the first
    time boundary and not the middle of the time interval.

    :param str filename: The filename
    :param re Object pattern: The filename pattern as a regex (from `re library \
    <https://docs.python.org/2/library/re.html>`_).
    :param str frequency: The time frequency
    :param str calendar: The NetCDF calendar attribute
    :returns: Start and end dates from the filename
    :rtype: *netcdftime.datetime*

    """
    dates = []
    date_as_since = None
    for key in ['period_start', 'period_end']:
        date = re.match(pattern, filename).groupdict()[key]
        digits = untruncated_timestamp(date)
        # Convert string digits to %Y-%m-%d %H:%M:%S format
        date_as_since = ''.join([''.join(triple) for triple in
                                 zip(digits[::2], digits[1::2], ['', '-', '-', ' ', ':', ':', ':'])])[:-1]
        # Use num2date to create netCDF4 datetime objects
        if frequency in ['3hr', '6hr']:
            # Fix on filename digits for 3hr and 6hr frequencies. 3hr (6hr) files always start
            # at 000000 end at 2100000 (180000) whether the time axis is instantaneous or not.
            dates.append(num2date(TIME_CORRECTION[frequency][key][digits[-6:]],
                                  units='days since ' + date_as_since,
                                  calendar=calendar))
        else:
            dates.append(num2date(0.0, units='days since ' + date_as_since, calendar=calendar))
    # Append date next to the end date for overlap diagnostic
    try:
        dates.append(num2date(time_inc(frequency)[0],
                              units=time_inc(frequency)[1] + ' since ' + date_as_since,
                              calendar=calendar)[0])
    except TypeError:
        dates.append(num2date(time_inc(frequency)[0],
                              units=time_inc(frequency)[1] + ' since ' + date_as_since,
                              calendar=calendar))
    return dates


def get_first_last_timesteps(ffp):
    """
    Returns first and last time steps from time axis of a NetCDF file.

    :param str ffp: The file full path
    :returns: The first and last timestep
    :rtype: *int*

    """
    try:
        nc = netCDF4.Dataset(ffp, 'r')
    except IOError:
        raise InvalidNetCDFFile(ffp)
    if 'time' not in nc.variables.keys():
        raise NoNetCDFVariable('time', ffp)
    first = nc.variables['time'][0]
    last = nc.variables['time'][-1]
    nc.close()
    return first, last


def get_next_timestep(ffp, current_timestep):
    """
    Returns next time step from time axis given the current one.

    :param str ffp: The file full path
    :param int current_timestep: The current_timestep
    :returns: The next timestep
    :rtype: *int*

    """
    try:
        nc = netCDF4.Dataset(ffp, 'r')
    except IOError:
        raise InvalidNetCDFFile(ffp)
    if 'time' not in nc.variables.keys():
        raise NoNetCDFVariable('time', ffp)
    try:
        index = int(np.where(nc.variables['time'][:] == current_timestep)[0][0])
    except IndexError:
        try:
            index = int(np.where(nc.variables['time'][:] == int(current_timestep))[0][0])
        except IndexError:
            raise NetCDFTimeStepNotFound(current_timestep, ffp)
    next_timestep = nc.variables['time'][index + 1]
    nc.close()
    return next_timestep


def time_inc(frequency):
    """
    Returns the time incrementation and time units depending on the MIP frequency.

    :param str frequency: The MIP frequency
    :returns: The corresponding time value and units
    :rtype: *list*

    """
    inc = {'subhr': [30, 'minutes'],
           '3hr': [3, 'hours'],
           '6hr': [6, 'hours'],
           'day': [1, 'days'],
           'mon': [1, 'months'],
           'yr': [1, 'years']}
    return inc[frequency]


def dates2int(dates):
    """
    Converts (a list of) dates as integers.

    :param list dates: A list of datetime or phony datetime objects
    :returns: The corresponding formatted integers
    :rtype: *list* or *int*

    """
    if isinstance(dates, list):
        return map(int, dates2str(dates, iso_format=False))
    else:
        return int(dates2str(dates, iso_format=False))


def dates2str(dates, iso_format=True):
    """
    Converts (a list of) dates in format: %Y%m%d %H:%M:%s.

    :param netcdftime.datetime/list dates: A list of datetime or phony datetime objects
    :param boolean iso_format: ISO format date if True
    :returns: The corresponding formatted strings
    :rtype: *list* or *str*

    """
    if isinstance(dates, list):
        return [date2str(date, iso_format) for date in dates]
    else:
        return date2str(dates, iso_format)


def date2str(date, iso_format=True):
    """
    Converts date in format: %Y%m%d %H:%M:%s.

    :param netcdftime.datetime date: A datetime or phony datetime objects
    :param boolean iso_format: ISO format date if True
    :returns: The corresponding formatted string
    :rtype: *str*

    """
    if iso_format:
        return '{0:04d}-{1:02d}-{2:02d}T{3:02d}:{4:02d}:{5:02d}'.format(date.year,
                                                                        date.month,
                                                                        date.day,
                                                                        date.hour,
                                                                        date.minute,
                                                                        date.second)
    else:
        return '{0:04d}{1:02d}{2:02d}{3:02d}{4:02d}{5:02d}'.format(date.year,
                                                                   date.month,
                                                                   date.day,
                                                                   date.hour,
                                                                   date.minute,
                                                                   date.second)


def ints2date(dates):
    """
    Converts (a list of) integer into corresponding datetime objects.

    :param list dates: A list of formatted integer
    :returns: The corresponding datetime objects
    :rtype: *list* or *netcdftime.datetime*

    """
    if isinstance(dates, list):
        dates = map(str, dates)
    else:
        dates = str(dates)
    return strs2date(dates, iso_format=False)


def strs2date(dates, iso_format=False):
    """
    Converts (a list of) formatted string %Y%m%d%H:%M:%s into corresponding datetime objects.

    :param list/str dates: A list of formatted dated
    :param boolean iso_format: ISO format date if True
    :returns: The corresponding datetime objects
    :rtype: *list* or *netcdftime.datetime*

    """
    if isinstance(dates, list):
        return [str2date(date, iso_format) for date in dates]
    else:
        return str2date(dates, iso_format)


def str2date(date, iso_format=True):
    """
    Converts a string format %Y%m%d%H:%M:%s into datetime object.

    :param str date: The formatted date
    :param boolean iso_format: ISO format date if True
    :returns: The corresponding datetime
    :rtype: *netcdftime.datetime*

    """
    if iso_format:
        pattern = re.compile(r'^(?P<year>\d{})-'
                             r'(?P<month>\d{2})-'
                             r'(?P<day>\d{2})T'
                             r'(?P<hour>\d{2}):'
                             r'(?P<minute>\d{2}):'
                             r'(?P<second>\d{2})$')
    else:
        pattern = re.compile(r'^(?P<year>\d{4})'
                             r'(?P<month>\d{2})'
                             r'(?P<day>\d{2})'
                             r'(?P<hour>\d{2})'
                             r'(?P<minute>\d{2})'
                             r'(?P<second>\d{2})$')
    attr = pattern.match(str(date)).groupdict()
    for k, v in attr.iteritems():
        attr[k] = int(v)
    return datetime(**attr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
   :platform: Unix
   :synopsis: Useful functions to use with this package.

"""

import logging
import os
import re
from datetime import datetime

from constants import *


class LogFilter(object):
    """
    Log filter with upper log level to use with the Python
    `logging <https://docs.python.org/2/library/logging.html>`_ module.

    """

    def __init__(self, level):
        self.level = level

    def filter(self, log_record):
        """
        Set the upper log level.

        """
        return log_record.levelno <= self.level


def init_logging(log, level='INFO'):
    """
    Initiates the logging configuration (output, date/message formatting).
    If a directory is submitted the logfile name is unique and formatted as follows:
    ``name-YYYYMMDD-HHMMSS-JOBID.log``If ``None`` the standard output is used.

    :param str log: The logfile directory.
    :param str level: The log level.

    """
    logname = 'nctime-{}-{}'.format(datetime(1, 1, 1)._to_real_datetime().now().strftime("%Y%m%d-%H%M%S"), os.getpid())
    formatter = logging.Formatter(fmt='%(levelname)-10s %(asctime)s %(message)s')
    if log:
        if not os.path.isdir(log):
            os.makedirs(log)
        logfile = os.path.join(log, logname)
    else:
        logfile = os.path.join(os.getcwd(), logname)
    logging.getLogger().setLevel(logging.DEBUG)
    if log:
        handler = logging.FileHandler(filename='{}.log'.format(logfile), delay=True)
    else:
        handler = logging.StreamHandler()
    handler.setLevel(logging.__dict__[level])
    handler.addFilter(LogFilter(logging.CRITICAL))
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)


def trunc(f, n):
    """
    Truncates a float f to n decimal places before rounding

    :param float f: The number to truncates
    :param int n: Decimal number to place before rounding
    :returns: The corresponding truncated number
    :rtype: *float*

    """
    slen = len('%.*f' % (n, f))
    return float(str(f)[:slen])


def cmd_exists(cmd):
    """
    Checks if a Shell command exists.

    :returns: True if exists.
    :rtype: *boolean*

    """
    return any(
        os.access(os.path.join(path, cmd), os.X_OK)
        for path in os.environ["PATH"].split(os.pathsep)
    )


def match(pattern, string, negative=False):
    """
    Validates a string against a regular expression.
    Only match at the beginning of the string.

    :param str pattern: The regular expression to match
    :param str string: The string to test
    :param boolean negative: True if negative matching (i.e., exclude the regex)
    :returns: True if it matches
    :rtype: *boolean*

    """
    if negative:
        return True if not re.search(pattern, string) else False
    else:
        return True if re.search(pattern, string) else False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Custom exceptions used in this module.

"""


class ChecksumFail(Exception):
    """
    Raised when no configuration file found.

    """

    def __init__(self, checksum_type, path):
        self.msg = "{} checksum failed.".format(checksum_type)
        self.msg += "\n<file: {}>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NetCDFVariableRemoveFail(Exception):
    """
    Raised when NetCDF variable removal fails.

    """

    def __init__(self, variable, path):
        self.msg = "Cannot remove variable: {}.".format(variable)
        self.msg += "\n<file: {}>".format(path)
        super(self.__class__, self).__init__(self.msg)


class NetCDFAttributeRemoveFail(Exception):
    """
    Raised when NetCDF attribute removal fails.

    """

    def __init__(self, attribute, path, variable=None):
        if variable:
            self.msg = "Cannot remove attribute: {}.".format(attribute)
            self.msg += "\n<variable: {}>".format(variable)
        else:
            self.msg = "Cannot remove global attribute: {}.".format(attribute)
        self.msg += "\n<file: {}>".format(path)
        super(self.__class__, self).__init__(self.msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
   :platform: Unix
   :synopsis: Insert time diagnostic to SQLite database.

"""

import sqlite3

from constants import *


def connect(db_file):
    """
    Open connexion to a database. The database is created if not exists.

    When a database is accessed by multiple connections, and one of the processes modifies the database, the SQLite
    database is locked until that transaction is committed. The timeout parameter specifies how long the connection
    should wait for the lock to go away until raising an exception. The default for the timeout parameter is 5.0 sec.

    More info here => http://www.sqlite.org/faq.html#q5

    We increase the timeout so we are able to use sqlite3 to run manual query without stopping the script.

    By default, sqlite is in autocommit mode but the sqlite3 python module is not in autocommit mode by default. We
    don't want autocommit mode, so we leave it at its default, which will result in a plain "BEGIN" statement.
    If you want autocommit mode, then set isolation_level to None.

    :param str db_file: The destination file database to connect
    :returns: The database connector
    :rtype: *sqlite3.connect*

    """
    connexion = sqlite3.connect(db_file, TIMEOUT)
    # By name columns indexing
    connexion.row_factory = sqlite3.Row
    return connexion


def close(connexion):
    """
    Close database connexion.

    :param sqlite3.connect connexion: A ``sqlite3.connect`` instance

    """
    if is_connected(connexion):
        connexion.close()


def is_connected(connexion):
    """
    Test database connexion.

    :param sqlite3.connect connexion: A ``sqlite3.connect`` instance
    :returns: True if connexion is already established with the database
    :rtype: *boolean*

    """
    if connexion is None:
        return False
    else:
        return True


def create(db_file):
    """
    Creates the destination database.

    :param str db_file: The destination file database to create

    """
    # Open database connection
    connexion = connect(db_file)
    # Create destination table
    connexion.execute(
            """
            CREATE TABLE IF NOT EXISTS time_axis
            (id INTEGER PRIMARY KEY,
            project TEXT,
            realm TEXT,
            frequency TEXT,
            freq_units TEXT,
            variable TEXT,
            filename TEXT,
            start TEXT,
            end TEXT,
            last TEXT,
            length INT,
            file_units TEXT,
            status TEXT,
            file_ref TEXT,
            ref_units TEXT,
            calendar TEXT,
            is_instant INT,
            has_bounds INT,
            new_checksum TEXT,
            full_path TEXT,
            creation_date TEXT)
            """
    )
    # Commit changes
    connexion.commit()
    # Close database connexion
    close(connexion)


def insert(db_file, data):
    """
    Insert time diagnostic to database.

    :param str db_file: The destination file database
    :param dict data: The serialized data to persist into the database

    """
    # Open database connection
    connexion = connect(db_file)
    # Insert a new entry
    connexion.execute(
            """
            INSERT INTO time_axis VALUES
            (NULL,
            :project,
            :realm,
            :frequency,
            :funits,
            :variable,
            :filename,
            :start_date,
            :end_date,
            :last_date,
            :length,
            :time_units,
            :status,
            :ref,
            :tunits,
            :calendar,
            :is_instant,
            :has_bounds,
            :new_checksum,
            :ffp,
            :creation_date)
            """, data
    )
    # Commit changes
    connexion.commit()
    # Close database connexion
    close(connexion)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Useful functions to collect data from directories.

"""

import os

from nctime.utils.misc import match


class Collector(object):
    """
    Base collector class to yield regular NetCDF files.

    :param str source: The list of sources to parse
    :param object data: Any object to attach to each collected data
    :returns: The data collector
    :rtype: *iter*

    """

    def __init__(self, source, data=None):
        self.source = source
        self.data = data
        self.FileFilter = Filter()
        assert os.path.isdir(self.source)

    def __iter__(self):
        for root, _, filenames in os.walk(self.source, followlinks=True):
            for filename in sorted(filenames):
                ffp = os.path.join(root, filename)
                if os.path.isfile(ffp) and self.FileFilter(filename):
                    yield self.attach(ffp)

    def __len__(self):
        """
        Returns collector length.

        :returns: The number of items in the collector.
        :rtype: *int*

        """
        s = 0
        for _ in self.__iter__():
            s += 1
        return s

    def attach(self, item):
        """
        Attach any object to the each collector item.

        :param str item: The collector item
        :returns: The collector item with the "data" object
        :rtype: *tuple*

        """
        return (item, self.data) if self.data else item

    def first(self):
        """

        :return:
        """
        return self.__iter__().next()


class Filter(dict):
    """
    Regex dictionary with a call method to evaluate a string against several regular expressions.
    The dictionary values are 2-tuples with the regular expression as a string and a boolean
    indicating to match (i.e., include) or non-match (i.e., exclude) the corresponding expression.

    """

    def __setitem__(self, key, value):
        # Assertions on filters values
        if isinstance(value, tuple):
            assert len(value) == 2
            assert isinstance(value[0], str)
            assert isinstance(value[1], bool)
        else:
            assert isinstance(value, str)
            # Set negative = False by default
            value = (value, False)
        dict.__setitem__(self, key, value)

    def __call__(self, string):
        return all([match(regex, string, negative=negative) for regex, negative in self.values()])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    :platform: Unix
    :synopsis: Processing context used in this module.

"""

import logging
import os
import sys
from multiprocessing.dummy import Pool as ThreadPool

from ESGConfigParser import SectionParser

from constants import *
from nctime.utils.collector import Collector
from nctime.utils.constants import *
from nctime.utils.custom_exceptions import *
from nctime.utils.misc import cmd_exists
from nctime.utils.time import TimeInit


class ProcessingContext(object):
    """
    Encapsulates the processing context/information for main process.

    :param ArgumentParser args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *ProcessingContext*

    """

    def __init__(self, args, directory):
        self.directory = directory
        self.config_dir = args.i
        self.write = args.write
        self.force = args.force
        self.verbose = args.v
        self.project = args.project
        self.tunits_default = None
        if self.project in DEFAULT_TIME_UNITS.keys():
            self.tunits_default = DEFAULT_TIME_UNITS[self.project]
        self.threads = args.max_threads
        self.db = None
        if args.db:
            self.db = os.path.realpath(args.db)
        self.scan_files = None
        self.status = []

    def __enter__(self):
        # Get checksum client
        self.checksum_client, self.checksum_type = self.get_checksum_client()
        # Init configuration parser
        self.cfg = SectionParser(section='project:{}'.format(self.project), directory=self.config_dir)
        self.pattern = self.cfg.translate('filename_format')
        # Init data collector
        self.sources = Collector(source=self.directory, data=self)
        # Init collector filter
        # Exclude hidden and/or non-NetCDF files
        self.sources.FileFilter['base_filter'] = ('^[!.].*\.nc$', True)
        # Exclude fixed frequency
        self.sources.FileFilter['frequency_filter'] = ('(_fx_|_fixed_|_fx.|_fixed.)', True)
        # Set driving time properties
        self.tinit = TimeInit(project=self.project, ref=self.sources.first()[0], tunits_default=self.tunits_default)
        # Init threads pool
        self.pool = ThreadPool(int(self.threads))
        return self

    def __exit__(self, *exc):
        # Close tread pool
        self.pool.close()
        self.pool.join()
        # Decline outputs depending on the scan results
        # Default is sys.exit(0)
        if any([s in EXIT_ERRORS for s in self.status]):
            logging.error('Some time axis should be corrected manually ({} files scanned)'.format(self.scan_files))
            sys.exit(1)
        else:
            logging.info('Time diagnostic completed ({} files scanned)'.format(self.scan_files))
            sys.exit(0)

    def get_checksum_client(self):
        """
        Gets the checksum client to use.
        Be careful to Exception constants by reading two different sections.

        :returns: The checksum client
        :rtype: *str*

        """
        _cfg = SectionParser(section='DEFAULT', directory=self.config_dir)
        if _cfg.has_option('DEFAULT', 'checksum'):
            checksum_client, checksum_type = _cfg.get_options_from_table('checksum')[0]
        else:  # Use SHA256 as default because esg.ini not mandatory in configuration directory
            checksum_client, checksum_type = 'sha256sum', 'SHA256'
        if not cmd_exists(checksum_client):
            raise ChecksumClientNotFound(checksum_client)
        else:
            return checksum_client, checksum_type
